/******************************************************************************
Nexus_Endpoint.C

Description:

  Classes to provide simple Nexus endpoint functionality

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "Nexus_internal.H"
#include "Nexus_Monitor.H"
#include "Nexus_Timer.H"

/******************************************************************************
			 Nexus_Endpoint handler table
******************************************************************************/



/******************************************************************************
			  Nexus_Endpoint methods
******************************************************************************/

/*
 * Nexus_Endpoint::Nexus_Endpoint()
 */
Nexus_Endpoint::Nexus_Endpoint()
{
    /*
     * Initialize concurrency control structures
     */
    nexus_mutex_init(&mutex, (nexus_mutexattr_t *) GLOBUS_NULL);
    nexus_cond_init(&cond, (nexus_condattr_t *) GLOBUS_NULL);

    /*
     * Initialize messsage queue structures
     */
    globus_fifo_init(&queue);

    /*
     * Initialize internal state
     */
    object_state = OBJECT_STATE_INACTIVE;
    pending_error = PENDING_ERROR_NONE;
    
    Nexus::status = Nexus::STATUS_SUCCESS;    
}
/* Nexus_Endpoint::Nexus_Endpoint() */


/*
 * Nexus_Endpoint::activate()
 */
Nexus_Bool
Nexus_Endpoint::activate(
    ARG_IN nexus_proto_type_t		proto_type,
    ARG_IN void *			proto_info_ptr,
    ARG_IN int				proto_info_size)
{
    static const nexus_handler_t
	Nexus_Endpoint_l_handler_table[] =
    {
	{NEXUS_HANDLER_TYPE_NON_THREADED,
	 Nexus_Endpoint::handle_Message},
	{NEXUS_HANDLER_TYPE_NON_THREADED,
	 (nexus_handler_func_t) NULL},
    };
    
    Nexus_Monitor			monitor(&mutex, &cond);

    Nexus::Status				status;
    
    /*
     * Verify that the object is in the correct state
     */
    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    /*
     * Create a endpoint attribute using the provided information
     */
    nexus_endpointattr_t		ep_attr;
    
    nexus_endpointattr_init(&ep_attr);

    nexus_endpointattr_set_handler_table(
	&ep_attr,
	(nexus_handler_t *) Nexus_Endpoint_l_handler_table,
	(sizeof(Nexus_Endpoint_l_handler_table)
	 / sizeof(nexus_handler_t)) - 1);

    nexus_endpointattr_set_unknown_handler(
	&ep_attr,
	Nexus_Endpoint::handle_Unknown,
	NEXUS_HANDLER_TYPE_NON_THREADED);

    /*
     * If the user requested a specific protocol to use, then attempt to use
     * that protocol.
     */
    int					rc;

    rc = NEXUS_SUCCESS;
    
    if (proto_type != NEXUS_PROTO_TYPE_ALL)
    {
	
	rc = nexus_endpointattr_set_protocol(&ep_attr,
					     proto_type,
					     proto_info_ptr,
					     proto_info_size);
	if (rc != NEXUS_SUCCESS)
	{
	    goto fn_abort;
	}
    }
    
    /*
     * Initialize the internal endpoint
     */
    rc = nexus_endpoint_init(&ep, &ep_attr);    
    if (rc != NEXUS_SUCCESS)
    {
	goto fn_abort;
    }

    nexus_endpoint_set_user_pointer(&ep, (void *) this);

  fn_abort:
    /*
     * Free up resources used by the endpoint attribute
     */
    nexus_endpointattr_destroy(&ep_attr);

    /*
     * Set the internal states and global status, and return
     */
    if (rc == NEXUS_SUCCESS)
    {
	object_state = OBJECT_STATE_ACTIVE;
	pending_error = PENDING_ERROR_NONE;
	Nexus::status = Nexus::STATUS_SUCCESS;

	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;

	return false;
    }
}
/* Nexus_Endpoint::activate() */


/*
 * Nexus_Endpoint::receive()
 */
Nexus_Bool
Nexus_Endpoint::receive(
    ARG_OUT Nexus_Receive_Buffer &	buffer,
    ARG_IN const long	 		timeout)
{
    Nexus_Monitor			monitor(&mutex, &cond);

    Nexus::Status				status;
    
    /*
     * Verify that the object is in the correct state
     */
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    Nexus_Timer				timer(timeout);
    
    /*
     * Wait for a message to arrive, the timer to expire, or for something to
     * go wrong
     */
    if (globus_fifo_empty(&queue))
    {
	do
	{
	    status = detect_Pending_Error();
	    if (status != Nexus::STATUS_SUCCESS)
	    {
		Nexus::status = status;
		pending_error = PENDING_ERROR_NONE;
		return false;
	    }
	    
	    monitor.wait(timer.get_Remaining_Time());

	    status = verify_Object_State(OBJECT_STATE_ACTIVE);
	    if (status != Nexus::STATUS_SUCCESS)
	    {
		Nexus::status = status;
		return false;
	    }
    
	    if (timer.get_Remaining_Time() == 0)
	    {
		Nexus::status = Nexus::STATUS_FAILURE;
		return false;
	    }
	}
	while(globus_fifo_empty(&queue));
    }

    /*
     * A new message arrived, so initialize the user's buffer with the message
     *
     * TODO: This code should be fixed once Nexus dispatch is fixed.  We are
     * making the assumption that nexus_buffer_t will fit in a (void *), which
     * it currently will, but may not in the future.
     */
    nexus_buffer_t			nx_buffer;

    nx_buffer = (nexus_buffer_t) globus_fifo_dequeue(&queue);
    buffer.activate(nx_buffer);
    
    Nexus::status = Nexus::STATUS_SUCCESS;

    return true;
}
/* Nexus_Endpoint::receive() */


/*
 * Nexus_Endpoint::poll()
 */
Nexus_Bool
Nexus_Endpoint::poll()
{
    nexus_poll();

    {
	Nexus_Monitor			monitor(&mutex);
	Nexus::Status			status;

	status = verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != Nexus::STATUS_SUCCESS)
	{
	    Nexus::status = status;
	    return false;
	}
	
	Nexus::status = Nexus::STATUS_SUCCESS;
	return (globus_fifo_empty(&queue) ? false : true);
    }
}
/* Nexus_Endpoint::poll() */


/*
 * Nexus_Endpoint::l_deactivate()
 */
Nexus_Bool
Nexus_Endpoint::l_deactivate()
{
    Nexus::Status			status;
    
    /*
     * Verify that we are in the correct state
     */
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
	
    }
    
    /*
     * Destroy the internal endpoint
     */
    nexus_endpoint_destroy(&ep);

    /*
     * Drain the queue
     */
    while(!globus_fifo_empty(&queue))
    {
	nexus_buffer_t			nx_buffer;
	
	nx_buffer = (nexus_buffer_t) globus_fifo_dequeue(&queue);
	nexus_buffer_destroy(&nx_buffer);
    }

    /*
     * Set the new object state
     */
    object_state = OBJECT_STATE_INACTIVE;

    /*
     * If there are any pending errors, then let the user know about them
     */
    status = detect_Pending_Error();
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	pending_error = PENDING_ERROR_NONE;
	return false;
	
    }

    return true;
}
/* Nexus_Endpoint::l_deactivate() */


/*
 * Nexus_Endpoint::deactivate(void)
 */
Nexus_Bool
Nexus_Endpoint::deactivate()
{
    Nexus_Monitor			monitor(&mutex, &cond);

    return l_deactivate();
}
/* Nexus_Endpoint::deactivate(void) */


/*
 * Nexus_Endpoint::~Nexus_Endpoint()
 */
Nexus_Endpoint::~Nexus_Endpoint()
{
    /*
     * Release the endpoint, if it is still active
     */
    {
	Nexus_Monitor			monitor(&mutex, &cond);

	l_deactivate();
	object_state = OBJECT_STATE_DELETED;
    }

    /*
     * Free resources used by message queue
     */
    globus_fifo_destroy(&queue);

    /*
     * Destroy concurrency control structures
     */
    nexus_cond_destroy(&cond);
    nexus_mutex_destroy(&mutex);

    object_state = OBJECT_STATE_DELETED;
    Nexus::status = Nexus::STATUS_SUCCESS;
}
/* Nexus_Endpoint::~Nexus_Endpoint() */


/*
 * Nexus_Endpoint::bind_Startpoint()
 */
Nexus_Bool
Nexus_Endpoint::bind_Startpoint(
    nexus_startpoint_t *		sp)
{
    Nexus_Monitor			monitor(&mutex, &cond);

    int					rc;
    Nexus::Status			status;

    /*
     * Verify that we have an active endpoint
     */
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
	
    }

    /*
     * Bind a new startpoint to the endpoint
     */
    rc = nexus_startpoint_bind(sp, &ep);

    if (rc == NEXUS_SUCCESS)
    {
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Endpoint::bind_Startpoint() */


/*
 * Nexus_Endpoint::handle_Message() 
 *
 * receive an incoming message and queue it
 *
 * Paramters:
 *
 *     ep - endpoint associated with the connection
 *
 *     buffer - buffer containing the message
 *
 *     hid - handler id
 *
 * Message Data:
 *
 *     char(*) - message data (completely defined by user)
 */
void
Nexus_Endpoint::handle_Message(
    nexus_endpoint_t *			ep,
    nexus_buffer_t *			buf,
    int					hid)
{
    Nexus_Endpoint *			endpoint;

    endpoint = (Nexus_Endpoint *) nexus_endpoint_get_user_pointer(ep);

    /*
     * Tell Nexus not to destroy this buffer
     */
    nexus_buffer_save(buf);
    
    /*
     * Add the buffer to the message queue and wakeup one of the waiting
     * receivers
     */
    {
	Nexus_Monitor			monitor(&endpoint->mutex,
						&endpoint->cond);

	/*
	 * TODO: This code should be fixed once Nexus dispatch is fixed.
	 * We are making the assumption that nexus_buffer_t will fit in
	 * a (void *), which it currently will, but may not in the future.
	 *
	 * Unfortunately, we had to do this because the nexus_buffer_t passed
	 * to use is allocated on the stack.
	 */
	globus_fifo_enqueue(&endpoint->queue, (void *) *buf);
	

	monitor.signal();
    }
}
/* Nexus_Endpoint::handle_Message() */


/*
 * Nexus_Endpoint::handle_Unknown()
 *
 * handle unknown message
 *
 * Paramters:
 *
 *     ep - endpoint associated with the connection
 *
 *     buffer - buffer containing the message
 *
 *     hid - bad handler id
 *
 * Message Data:
 *
 *     none
 */
void
Nexus_Endpoint::handle_Unknown(
    ARG_IN nexus_endpoint_t *		ep,
    ARG_INMOD nexus_buffer_t *		buf,
    ARG_IN int				hid)
{
    Nexus_Endpoint *			endpoint;
    
    /*
     * If we are debugging, then this is definitely an error we wan to catch
     */
    NexusAssert(NEXUS_FALSE);

    endpoint = (Nexus_Endpoint *) nexus_endpoint_get_user_pointer(ep);

    /*
     * However, if we are running production code, we should just return an
     * error and hope things don't fall over completely...
     */
    {
	Nexus_Monitor			monitor(&endpoint->mutex,
						&endpoint->cond);

	endpoint->pending_error = PENDING_ERROR_PROTO;

	monitor.broadcast();
    }
}
/* Nexus_Endpoint::handle_Unknown() */


/*
 * Nexus_Endpoint::verify_Object_State()
 */
Nexus::Status
Nexus_Endpoint::verify_Object_State(
    Object_State			desired_state)
{
    if (object_state == desired_state)
    {
	return Nexus::STATUS_SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (object_state)
    {
      case OBJECT_STATE_DELETED:
	NexusAssert(object_state != OBJECT_STATE_DELETED);
	return Nexus::STATUS_OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return Nexus::STATUS_OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return Nexus::STATUS_OBJECT_INACTIVE;

      default:
	return Nexus::STATUS_INTERNAL_ERROR;
    }
}
/* Nexus_Endpoint::verify_Object_State() */


/*
 * Nexus_Endpoint::detect_Pending_Error()
 */
Nexus::Status
Nexus_Endpoint::detect_Pending_Error()
{
    switch (pending_error)
    {
      case PENDING_ERROR_NONE:
	return Nexus::STATUS_SUCCESS;
	    
      case PENDING_ERROR_MEM:
	/*
	 * An out of memory error may not be a permanent situation.  Tell
	 * the user once that the error occurred so he know that he may
	 * have lost some messages, but then reset the state so he can
	 * continue to receive any new messages.
	 */
	return Nexus::STATUS_INSUFFICIENT_MEMORY;

      case PENDING_ERROR_PROTO:
      default:
	return Nexus::STATUS_INTERNAL_ERROR;
    }
}
/* Nexus_Endpoint::detect_Pending_Error() */
