/******************************************************************************
Nexus_Listener.C

Description:

  Classes to provide simple Nexus functionality

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "Nexus_internal.H"


/******************************************************************************
			    Nexus_Listener methods
******************************************************************************/

/*
 * Nexus_Listener::Nexus_Listener()
 */
Nexus_Listener::Nexus_Listener()
{
    /*
     * Initialize concurrency control structures
     */
    nexus_mutex_init(&mutex, (nexus_mutexattr_t *) GLOBUS_NULL);

    object_state = OBJECT_STATE_INACTIVE;
}

/*
 * Nexus_Listener::activate()
 */
Nexus_Bool
Nexus_Listener::activate(
    ARG_IN Nexus_Endpoint &		endpoint_arg,
    ARG_IN unsigned short		port_arg)
{
    Nexus_Monitor			monitor(&mutex);
    
    int					rc;

    /*
     * Verify that we are operating on an inactive but valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    endpoint = &endpoint_arg;
    port = port_arg;
    
    rc = nexus_allow_attach(&port,
			    &hostname,
			    Nexus_Listener::handle_Attach,
			    (void *) this);

    if (rc == NEXUS_SUCCESS)
    {
	object_state = OBJECT_STATE_ACTIVE;
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Listener::Nexus_Listener() */
    

/*
 * Nexus_Listener::get_Hostname()
 */
char *
Nexus_Listener::get_Hostname()
{
    Nexus_Monitor			monitor(&mutex);
    
    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return GLOBUS_NULL;
    }
    
    Nexus::status = Nexus::STATUS_SUCCESS;
    return hostname;
}
/* Nexus_Listener::get_Hostname() */

    
/*
 * Nexus_Listener::get_Port()
 */
unsigned short
Nexus_Listener::get_Port()
{
    Nexus_Monitor			monitor(&mutex);
    
    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return 0;
    }
    
    Nexus::status = Nexus::STATUS_SUCCESS;
    return port;
}
/* Nexus_Listener::get_Port() */


/*
 * Nexus_Listener::l_deactivate()
 */
Nexus_Bool
Nexus_Listener::l_deactivate()
{
    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    /*
     * Disallow furter attach requests
     */
    nexus_disallow_attach(port);

    object_state = OBJECT_STATE_INACTIVE;
    Nexus::status = Nexus::STATUS_SUCCESS;
    return true;
}
/* Nexus_Listener::l_deactivate() */

    
/*
 * Nexus_Listener::deactivate(void)
 */
Nexus_Bool
Nexus_Listener::deactivate()
{
    Nexus_Monitor			monitor(&mutex);

    return l_deactivate();
}
/* Nexus_Listener::deactivate(void) */

    
/*
 * Nexus_Listener::~Nexus_Listener()
 */
Nexus_Listener::~Nexus_Listener()
{
    /*
     * Disallow further attach requests
     */
    {
	Nexus_Monitor			monitor(&mutex);

	l_deactivate();
	object_state = OBJECT_STATE_DELETED;
    }
    
    /*
     * Destroy concurrency control structures
     */
    nexus_mutex_destroy(&mutex);
    
    Nexus::status = Nexus::STATUS_SUCCESS;
}
/* Nexus_Listener::~Nexus_Listener() */

    
/*
 * Nexus_Listener::verify_Object_State()
 */
Nexus::Status
Nexus_Listener::verify_Object_State(
    Object_State			desired_state)
{
    if (object_state == desired_state)
    {
	return Nexus::STATUS_SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (object_state)
    {
      case OBJECT_STATE_DELETED:
	NexusAssert(object_state != OBJECT_STATE_DELETED);
	return Nexus::STATUS_OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return Nexus::STATUS_OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return Nexus::STATUS_OBJECT_INACTIVE;

      default:
	return Nexus::STATUS_INTERNAL_ERROR;
    }
}
/* Nexus_Listener::verify_Object_State() */


/*
 * Nexus_Listener::handle_Attach()
 *
 * Handle attach requests by sending the remote process a startpoint to this
 * listener's endpoint
 *
 * Paramters:
 *
 *     arg - pointer to listener class
 *
 *     my_url - the URL of this listener
 *
 *     sp_ptr - pointer to a startpoint (which needs to be filled in)
 *
 */
int
Nexus_Listener::handle_Attach(
    void *				arg,
    char *				my_url,
    nexus_startpoint_t *		sp_ptr)
{
    Nexus_Listener *			listener;

    listener = (Nexus_Listener *) arg;

    Nexus_Monitor			monitor(&listener->mutex);
    
    if ((listener->verify_Object_State(OBJECT_STATE_ACTIVE)
	 == Nexus::STATUS_SUCCESS)
	&& listener->endpoint->bind_Startpoint(sp_ptr))
    {
	return NEXUS_SUCCESS;
    }
    else
    {
	return NEXUS_ERROR_ACCEPT_ATTACH_FAILED;
    }
}
/* Nexus_Listener::handle_Attach() */
