/******************************************************************************
Nexus_Startpoint.C

Description:

  Classes to provide simple Nexus startpoint functionality

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "Nexus_internal.H"
#include "Nexus_Monitor.H"

#include <stdio.h>


/******************************************************************************
			   Nexus_Startpoint methods
******************************************************************************/


/*
 * Nexus_Startpoint::Nexus_Startpoint()
 */
Nexus_Startpoint::Nexus_Startpoint()
{
    /*
     * Initialize concurrency control structures
     */
    nexus_mutex_init(&mutex, (nexus_mutexattr_t *) GLOBUS_NULL);

    /*
     * Set state to indicate that the object has been successfully initialized
     */
    object_state = OBJECT_STATE_INACTIVE;
    Nexus::status = Nexus::STATUS_SUCCESS;
}
/* Nexus_Startpoint::Nexus_Starpoint() */


/*
 * Nexus_Startpoint::activate(host, port)
 */
Nexus_Bool
Nexus_Startpoint::activate(
    ARG_IN char *			host,
    ARG_IN unsigned short		port)
{
    int					rc;
    char *				url;

    /*
     * Verify that we are operating on a valid but inactive object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    /*
     * Obtain a new startpoint by attaching to a remote endpoint
     */
    url = new char [strlen(host) + 22];
    sprintf(url, "x-nexus://%s:%d/", host, port);

    rc = nexus_attach(url, &sp);
    delete [] url;

    if (rc == NEXUS_SUCCESS)
    {
	object_state = OBJECT_STATE_ACTIVE;
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::activate(host, port) */


/*
 * Nexus_Startpoint::activate(endpoint)
 */
Nexus_Bool
Nexus_Startpoint::activate(
    ARG_IN Nexus_Endpoint &		endpoint)
{
    /*
     * Verify that we are operating on a valid but inactive object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }

    /*
     * Create the startpoint from a local endpoint
     */
    endpoint.bind_Startpoint(&sp);
    if (Nexus::status == Nexus::STATUS_SUCCESS)
    {
	object_state = OBJECT_STATE_ACTIVE;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::activate(endpoint) */


/*
 * Nexus_Startpoint::send()
 */
Nexus_Bool
Nexus_Startpoint::send(
    ARG_INOUT Nexus_Send_Buffer &	buffer)
{
    Nexus_Monitor			monitor(&mutex);

    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }

    /*
     * Send the message
     */
    if (buffer.send(&sp, Nexus_Endpoint_l_handle_Message_HID))
    {
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	/* XXX: need to set error code */
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::send() */


/*
 * Nexus_Startpoint::l_deactivate()
 */
Nexus_Bool
Nexus_Startpoint::l_deactivate()
{
    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }

    /*
     * Set the object state to inactive
     */
    object_state = OBJECT_STATE_INACTIVE;
    
    /*
     * Destroy the startpoint, freeing resources held by it
     */
    int					rc;
    
    rc = nexus_startpoint_destroy(&sp);

    if (rc == NEXUS_SUCCESS)
    {
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::l_deactivate() */


/*
 * Nexus_Startpoint::deactivate()
 */
Nexus_Bool
Nexus_Startpoint::deactivate()
{
    Nexus_Monitor			monitor(&mutex);

    return l_deactivate();
}
/* Nexus_Startpoint::deactivate() */


/*
 * Nexus_Startpoint::~Nexus_Startpoint()
 */
Nexus_Startpoint::~Nexus_Startpoint()
{
    /*
     * First, deactivate the object
     */
    {
	Nexus_Monitor			monitor(&mutex);

	l_deactivate();
	object_state = OBJECT_STATE_DELETED;
    }

    /*
     * Destroy concurrency control structures
     */
    nexus_mutex_destroy(&mutex);

    /*
     * Set state to indicate that the object has been deleted
     */
    object_state = OBJECT_STATE_DELETED;
    
    Nexus::status = Nexus::STATUS_SUCCESS;
}
/* Nexus_Startpoint::~Nexus_Startpoint() */


/*
 * Nexus_Startpoint::sizeof_Internal_Startpoint()
 */
Nexus_Size
Nexus_Startpoint::sizeof_Internal_Startpoint()
{
    Nexus_Monitor			monitor(&mutex);

    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return 0;
    }

    /*
     * Get the sizeof the internal startpoint
     */
    Nexus_Size				size;
    
    size = (Nexus_Size) nexus_sizeof_startpoint(&sp, 1);
    if (size > 0)
    {
	Nexus::status = Nexus::STATUS_SUCCESS;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
    }
	
    return size;
}
/* Nexus_Startpoint::put_Internal_Startpoint() */


/*
 * Nexus_Startpoint::get_Internal_Startpoint()
 */
Nexus_Bool
Nexus_Startpoint::get_Internal_Startpoint(
    ARG_INOUT nexus_buffer_t *		nx_buffer)
{
    Nexus_Monitor			monitor(&mutex);

    /*
     * Verify that we are operating on a valid but inactive object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }
    
    /*
     * Get the startpoint out of the buffer
     */
    int					rc;

    rc = nexus_get_startpoint(nx_buffer, &sp, 1);
    if (rc == NEXUS_SUCCESS)
    {
	object_state = OBJECT_STATE_ACTIVE;
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::get_Internal_Startpoint() */


/*
 * Nexus_Startpoint::put_Internal_Startpoint()
 */
Nexus_Bool
Nexus_Startpoint::put_Internal_Startpoint(
    ARG_INOUT nexus_buffer_t *		nx_buffer)
{
    Nexus_Monitor			monitor(&mutex);

    /*
     * Verify that we are operating on a valid object
     */
    Nexus::Status			status;
    
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != Nexus::STATUS_SUCCESS)
    {
	Nexus::status = status;
	return false;
    }

    /*
     * Transfer the startpoint into the buffer
     */
    int					rc;
    
    rc = nexus_put_startpoint_transfer(nx_buffer, &sp, 1);
    if (rc == NEXUS_SUCCESS)
    {
	object_state = OBJECT_STATE_INACTIVE;
	Nexus::status = Nexus::STATUS_SUCCESS;
	return true;
    }
    else
    {
	Nexus::status = Nexus::STATUS_FAILURE;
	return false;
    }
}
/* Nexus_Startpoint::put_Internal_Startpoint() */


/*
 * Nexus_Startpoint::verify_Object_State()
 */
Nexus::Status
Nexus_Startpoint::verify_Object_State(
    ARG_IN Object_State			desired_state)
{
    if (object_state == desired_state)
    {
	return Nexus::STATUS_SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (object_state)
    {
      case OBJECT_STATE_DELETED:
	NexusAssert(object_state != OBJECT_STATE_DELETED);
	return Nexus::STATUS_OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return Nexus::STATUS_OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return Nexus::STATUS_OBJECT_INACTIVE;

      default:
	return Nexus::STATUS_INTERNAL_ERROR;
    }
}
/* Nexus_Startpoint::verify_Object_State() */


